"""Definitions for the `BNSEjecta` class."""
# import astropy.constants as c
import numpy as np
from astrocats.catalog.source import SOURCE

from mosfit.constants import FOE, KM_CGS, M_SUN_CGS, C_CGS, G_CGS
from mosfit.modules.energetics.energetic import Energetic


# G_CGS = c.G.cgs.value


class BNSEjecta(Energetic):
    """
    Generate `mejecta`, `vejecta` and `kappa` from neutron star binary
    parameters.

    Includes tidal and shocked dynamical and disk wind ejecta following
    Dietrich+ 2017 and Coughlin+ 2019, with opacities from Sekiguchi+ 2016,
    Tanaka+ 2019, Metzger and Fernandez 2014, Lippuner+ 2017

    Also includes an ignorance parameter `alpha` for NS-driven winds to
    increase the fraction of blue ejecta: Mdyn_blue /= alpha
     - therefore NS surface winds turned off by setting alpha = 1
    """

    _REFERENCES = [
        {SOURCE.BIBCODE: '2017CQGra..34j5014D'},
        {SOURCE.BIBCODE: '2019MNRAS.489L..91C'},
        {SOURCE.BIBCODE: '2013PhRvD..88b3007M'},
        {SOURCE.BIBCODE: '2016PhRvD..93l4046S'},
        {SOURCE.BIBCODE: '2014MNRAS.441.3444M'},
        {SOURCE.BIBCODE: '2017MNRAS.472..904L'},
        {SOURCE.BIBCODE: '2019LRR....23....1M'},
        {SOURCE.BIBCODE: '2020MNRAS.496.1369T'},
        {SOURCE.BIBCODE: '2018PhRvL.121i1102D'}
    ]

    def process(self, **kwargs):
        """Process module."""
        ckm = C_CGS / KM_CGS
        self._mchirp = kwargs[self.key('Mchirp')]
        self._q = kwargs[self.key('q')]
        # Mass of heavier NS
        self._m1 = self._mchirp * self._q**-0.6 * (self._q+1)**0.2
        # Mass of lighter NS
        self._m2 = self._m1*self._q
        self._m_total = self._m1 + self._m2
        # How much of disk is ejected
        self._disk_frac = kwargs[self.key('disk_frac')]
        # Max mass for non-rotating NS
        self._m_tov = kwargs[self.key('Mtov')]
        # NS radius
        self._radius_ns = kwargs[self.key('radius_ns')]
        # Fraction of blue ejecta from dynamical shocks (Coughlin+ 2019)
        # Here we are assuming remainder is a NS wind
        # So only applicable if merger product avoids prompt collapse
        self._alpha = kwargs[self.key('alpha')]
        # Opening angle
        self._cos_theta_open = kwargs[self.key('cos_theta_open')]

        theta_open = np.arccos(self._cos_theta_open)

        # Additional systematic scatter (if desired)
        self._errMdyn = kwargs[self.key('errMdyn')]
        self._errMdisk = kwargs[self.key('errMdisk')]


        C1 = G_CGS * self._m1 * M_SUN_CGS /(self._radius_ns*1e5 * C_CGS**2)
        C2 = G_CGS * self._m2 * M_SUN_CGS /(self._radius_ns*1e5 * C_CGS**2)

        # Baryonic masses, Gao 2019
        Mb1 = self._m1 + 0.08*self._m1**2
        Mb2 = self._m2 + 0.08*self._m2**2


        # Dynamical ejecta:

        # Fitting function from Dietrich and Ujevic 2017
        a_1 = -1.35695
        b_1 = 6.11252
        c_1 = -49.43355
        d_1 = 16.1144
        n = -2.5484
        Mejdyn = 1e-3* (a_1*((self._m2/self._m1)**(1/3)*(1-2*C1)/C1*Mb1 +
                        (self._m1/self._m2)**(1/3)*(1-2*C2)/C2*Mb2) +
                        b_1*((self._m2/self._m1)**n*Mb1 + (self._m1/self._m2)**n*Mb2) +
                        c_1*(Mb1-self._m1 + Mb2-self._m2) + d_1)

        Mejdyn *= self._errMdyn

        if Mejdyn < 0:
            Mejdyn = 0

        # Calculate fraction of ejecta with Ye<0.25 from fits to Sekiguchi 2016
        # Also consistent with Dietrich: mostly blue at M1/M2=1, all red by M1/M2=1.2.
        # And see Bauswein 2013, shocked (blue) component decreases with M1/M2
        a_4 = 14.8609
        b_4 = -28.6148
        c_4 = 13.9597

        f_red = min([a_4*(self._m1/self._m2)**2+b_4*(self._m1/self._m2)+c_4,1]) # fraction can't exceed 100%


        # Velocity of dynamical ejecta
        a_2 = -0.219479
        b_2 = 0.444836
        c_2 = -2.67385

        vdynp = a_2*((self._m1/self._m2)*(1+c_2*C1) + (self._m2/self._m1)*(1+c_2*C2)) + b_2


        a_3 = -0.315585
        b_3 = 0.63808
        c_3 = -1.00757

        vdynz = a_3*((self._m1/self._m2)*(1+c_3*C1) + (self._m2/self._m1)*(1+c_3*C2)) + b_3

        vdyn = np.sqrt(vdynp**2+vdynz**2)

        # average velocity over angular ranges (< and > theta_open)

        theta1 = np.arange(0,theta_open,0.01)
        theta2 = np.arange(theta_open,np.pi/2,0.01)

        vtheta1 = np.sqrt((vdynz*np.cos(theta1))**2+(vdynp*np.sin(theta1))**2)
        vtheta2 = np.sqrt((vdynz*np.cos(theta2))**2+(vdynp*np.sin(theta2))**2)

        atheta1 = 2*np.pi*np.sin(theta1)
        atheta2 = 2*np.pi*np.sin(theta2)

        vejecta_blue = np.trapz(vtheta1*atheta1,x=theta1)/np.trapz(atheta1,x=theta1)
        vejecta_red = np.trapz(vtheta2*atheta2,x=theta2)/np.trapz(atheta2,x=theta2)

        mejecta_red = Mejdyn * f_red
        vejecta_red *= ckm

        mejecta_blue = Mejdyn * (1-f_red)
        vejecta_blue *= ckm


        # Bauswein 2013, cut-off for prompt collapse to BH
        Mthr = (2.38-3.606*self._m_tov/self._radius_ns)*self._m_tov

        if self._m_total < Mthr:
            mejecta_blue /= self._alpha


        # Now compute disk ejecta following Coughlin+ 2019

        a_5 = -31.335
        b_5 = -0.9760
        c_5 = 1.0474
        d_5 = 0.05957

        logMdisk = np.max([-3, a_5*(1+b_5*np.tanh((c_5-self._m_total/Mthr)/d_5))])

        Mdisk = 10**logMdisk

        Mdisk *= self._errMdisk
        
        Mejdisk = Mdisk * self._disk_frac
        mejecta_purple = Mejdisk


        # Fit for disk velocity using Metzger and Fernandez
        vdisk_max = 0.15
        vdisk_min = 0.03
        vfit = np.polyfit([self._m_tov,Mthr],[vdisk_max,vdisk_min],deg=1)

        # Get average opacity of 'purple' (disk) component
        # Mass-averaged Ye as a function of remnant lifetime from Lippuner 2017
        # Lifetime related to Mtot using Metzger handbook table 3
        if self._m_total < self._m_tov:
            # stable NS
            Ye = 0.38
            vdisk = vdisk_max
        elif self._m_total < 1.2*self._m_tov:
            # long-lived (>>100 ms) NS remnant Ye = 0.34-0.38,
            # smooth interpolation
            Yfit = np.polyfit([self._m_tov,1.2*self._m_tov],[0.38,0.34],deg=1)
            Ye = Yfit[0]*self._m_total + Yfit[1]
            vdisk = vfit[0]*self._m_total + vfit[1]
        elif self._m_total < Mthr:
            # short-lived (hypermassive) NS, Ye = 0.25-0.34, smooth interpolation
            Yfit = np.polyfit([1.2*self._m_tov,Mthr],[0.34,0.25],deg=1)
            Ye = Yfit[0]*self._m_total + Yfit[1]
            vdisk = vfit[0]*self._m_total + vfit[1]
        else:
            # prompt collapse to BH, disk is red
            Ye = 0.25
            vdisk = vdisk_min

        # Convert Ye to opacity using Tanaka et al 2019 for Ye >= 0.25:
        a_6 = 2112.0
        b_6 = -2238.9
        c_6 = 742.35
        d_6 = -73.14

        kappa_purple = a_6*Ye**3 + b_6*Ye**2 + c_6*Ye + d_6

        vejecta_purple = vdisk * ckm


        return {self.key('mejecta_blue'): mejecta_blue,
                self.key('mejecta_red'): mejecta_red,
                self.key('mejecta_purple'): mejecta_purple,
                self.key('mejecta_dyn'): Mejdyn,
                self.key('vejecta_blue'): vejecta_blue,
                self.key('vejecta_red'): vejecta_red,
                self.key('vejecta_purple'): vejecta_purple,
                self.key('kappa_purple'): kappa_purple,
                self.key('M1'): self._m1,
                self.key('M2'): self._m2,
                self.key('radius_ns'): self._radius_ns
                }



"""Definitions for the `Shock` class."""
from math import isnan
from astrocats.catalog.source import SOURCE

import numpy as np

from mosfit.constants import C_CGS, DAY_CGS, FOUR_PI, KM_CGS, M_SUN_CGS
from mosfit.modules.engines.engine import Engine


# Important: Only define one ``Module`` class per file.


class Shock(Engine):
    """Cooling emission from shock-heated cocoon.

    Follows Piro and Kollmeier 2018.

    Shock heating can be turned off in bns model by setting cos_theta_cocoon=1
    """

    DIFF_CONST = M_SUN_CGS / (FOUR_PI * C_CGS * KM_CGS)

    _REFERENCES = [
        {SOURCE.BIBCODE: '2018ApJ...855..103P'}
    ]

    def process(self, **kwargs):
        """Process module."""
        self._times = kwargs[self.key('dense_times')]
        self._rest_t_explosion = kwargs[self.key('resttexplosion')]
        self._kappa = kwargs[self.key('kappa')]
        self._m_ejecta = kwargs[self.key('mejecta')]
        self._v_ejecta = kwargs[self.key('vejecta')]
        # Cocoon opening angle- should be between 1 and cos_theta_open
        # for polar KN ejecta
        self._cos_theta_cocoon = kwargs[self.key('cos_theta_cocoon')]
        # Shocked ejecta power law density profile
        self._s =  kwargs[self.key('s')]
        # Shock breakout timescale in seconds
        self._t_shock =  kwargs[self.key('tshock')]
        # Radius where material is shocked by relativistic jet:
        R = C_CGS * self._t_shock

        theta = np.arccos(self._cos_theta_cocoon)

        self._tau_diff = np.sqrt(self.DIFF_CONST * self._kappa *
                                 self._m_ejecta / self._v_ejecta) / DAY_CGS


        L0 = (theta**2/2)**(1/3) * (self._m_ejecta * M_SUN_CGS *
                self._v_ejecta * KM_CGS * R / (self._tau_diff * DAY_CGS)**2 )

        ts = [
            np.inf
            if self._rest_t_explosion > x else (x - self._rest_t_explosion)
            for x in self._times
        ]

        luminosities = [L0 * (t/self._tau_diff)**-(4/(self._s+2)) for t in ts]
        luminosities = [0.0 if isnan(x) else x for x in luminosities]

        return {self.dense_key('luminosities'): luminosities}



"""Definitions for the `DiffusionCSM` class."""
from collections import OrderedDict

import numpy as np
from scipy.interpolate import interp1d

from mosfit.constants import C_CGS, DAY_CGS, M_SUN_CGS, AU_CGS
from mosfit.modules.transforms.transform import Transform


# Important: Only define one ``Module`` class per file.


class DiffusionCSM(Transform):
    """Photon diffusion transform for CSM model."""

    N_INT_TIMES = 3000
    MIN_LOG_SPACING = -3


    def process(self, **kwargs):
        """Process module."""
        Transform.process(self, **kwargs)
        self._kappa = kwargs[self.key('kappa')]
        self._mass = kwargs[self.key('mcsm')] * M_SUN_CGS
        self._R0 = kwargs[self.key('r0')] * AU_CGS  # AU to cm
        self._s = kwargs[self.key('s')]
        self._rho = kwargs[self.key('rho')]
        self._mejecta = kwargs[self.key('mejecta')] * M_SUN_CGS  # Msol to grms

        # scaling constant for CSM density profile
        self._q = self._rho * self._R0 ** self._s
        # outer radius of CSM shell
        self._Rcsm = (
            (3.0 - self._s) / (4.0 * np.pi * self._q) * self._mass +
            self._R0 ** (3.0 - self._s)) ** (1.0 / (3.0 - self._s))
        # radius of photosphere (should be within CSM)
        self._Rph = abs(
            (-2.0 * (1.0 - self._s) / (3.0 * self._kappa * self._q) +
             self._Rcsm ** (1.0 - self._s)) ** (1.0 / (1.0 - self._s)))
        self._tau_diff = (
            self._kappa * self._mass) / (13.8 * C_CGS * self._Rph) / DAY_CGS

        # mass of the optically thick CSM (tau > 2/3).
        self._Mcsm_th = np.abs(4.0 * np.pi * self._q / (3.0 - self._s) * (
            self._Rph**(3.0 - self._s) - self._R0 **
            (3.0 - self._s)))
        beta =  4. * np.pi ** 3. / 9.
        td2 = self._tau_diff**2
        td = self._tau_diff
        t0 = self._kappa * (self._Mcsm_th) \
                / (beta * C_CGS * self._Rph) / DAY_CGS
        new_lums = np.zeros_like(self._times_to_process)
        if len(self._dense_times_since_exp) < 2:
            return {self.dense_key('luminosities'): new_lums}
        min_te = min(self._dense_times_since_exp)
        tb = max(0.0, min_te)
        linterp = interp1d(
            self._dense_times_since_exp, self._dense_luminosities, copy=False,
            assume_sorted=True, bounds_error=False, fill_value=0.0)

        uniq_times = np.unique(self._times_to_process[
            (self._times_to_process >= tb) & (
                self._times_to_process <= self._dense_times_since_exp[-1])])
        lu = len(uniq_times)

        num = int(round(self.N_INT_TIMES / 2.0))
        lsp = np.logspace(
            np.log10(t0 /
                     self._dense_times_since_exp[-1]) +
            self.MIN_LOG_SPACING, 0, num)
        xm = np.unique(np.concatenate((lsp, 1 - lsp)))

        int_times = np.clip(
            tb + (uniq_times.reshape(lu, 1) - tb) * xm, tb,
            self._dense_times_since_exp[-1])
        int_times = tb + (uniq_times.reshape(lu, 1) - tb) * xm
        int_tes = int_times[:, -1]

        int_lums = linterp(int_times)  # noqa: F841
        int_args = int_lums * np.exp((int_times) / t0)
        int_args[np.isnan(int_args)] = 0.0

        uniq_lums = np.trapz(int_args, int_times)
        uniq_lums*= np.exp(-int_tes/t0)/t0
        new_lums = uniq_lums[np.searchsorted(uniq_times,
                                             self._times_to_process)]

        return {self.dense_key('luminosities'): new_lums}



"""Definitions for the `Kroupa` class."""
import numpy as np
from mosfit.modules.parameters.parameter import Parameter


# Important: Only define one ``Module`` class per file.


class Kroupa(Parameter):
    """Kroupa IMF from https://arxiv.org/pdf/astro-ph/0102155.pdf.

    Defined for stellar masses larger than 0.01 solar masses.
    """

    def __init__(self, **kwargs):
        """Initialize module."""
        super(Kroupa, self).__init__(**kwargs)

        if self._min_value > self._max_value:  # min mass must be < max mass
            raise ValueError(
                'Star mass prior must have max value > min value')
        if self._min_value < 0.01:
            raise ValueError(
                'Star mass kroupa prior is only defined down ' +
                'to 0.01 solar masses. Edit prior range or change prior.')

        self._norm = 1. / self.kroupa_cdf(self._min_value, self._max_value, 1)

    def lnprior_pdf(self, x):
        """Evaluate natural log of probability density function.

        PDF from Kroupa (2001b)
        """
        value = self.value(x)

        if value < 0.08:
            return np.log((value / 0.08)**(-0.3))
        elif value < 0.5:
            return np.log((value / 0.08)**(-1.3))
        else:
            return np.log((0.5 / 0.08)**(-1.3) * (
                value / 0.5)**(-2.3))

    def kroupa_cdf(self, minmass, maxmass, k):
        """Cumulative density function from Kroupa 2001b.

        0.01 <= mass, k is normalization
        """
        if minmass > maxmass:
            return 0  # useful default for calculating iCDF below
        elif minmass < 0.08:
            if maxmass < 0.08:
                prob = (k / 0.7 * 0.08**0.3 * (maxmass**0.7 - minmass**0.7))
            elif maxmass < 0.5:
                prob = ((k / 0.7 * 0.08**0.3 * (0.08**0.7 - minmass**0.7)) +
                        (k / -0.3) * 0.08**1.3 * (maxmass**(-0.3) -
                                                  0.08**(-0.3)))
            else:
                prob = ((k / 0.7 * 0.08**0.3 * (0.08**0.7 - minmass**0.7)) +
                        (k / -0.3) * 0.08**1.3 * (0.5**(-0.3) - 0.08**(-0.3)) +
                        (k / -1.3) * 0.08**1.3 * (0.5 * maxmass**(-1.3) -
                                                  0.5**(-0.3)))

        elif 0.08 <= minmass < 0.5:
            if maxmass < 0.5:
                prob = ((k / -0.3) * 0.08**1.3 * (maxmass**(-0.3) -
                        minmass**(-0.3)))
            else:  # maxmass > 0.5
                prob = (((k / -0.3) * 0.08**1.3 * (0.5**(-0.3) -
                        minmass**(-0.3))) +
                        ((k / -1.3) * 0.08**1.3 * 0.5 *
                        (maxmass**(-1.3) - 0.5**(-1.3))))

        elif 0.5 <= minmass:
            prob = ((k / -1.3) * 0.08**1.3 * 0.5 *
                    (maxmass**(-1.3) - minmass**(-1.3)))

        return prob

    def prior_icdf(self, u):
        """Evaluate inverse cumulative density function from Kroupa 2001b.

        output mass scaled to 0-1 interval
        """
        #  u < cdf(value = 0.08)
        if u < self.kroupa_cdf(self._min_value, 0.08, self._norm):  # if minvalue > maxvalue, CDF returns 0
            value = (u * 0.7 / self._norm * 0.08**(-0.3) +
                     self._min_value**0.7)**(1.0 / 0.7)
        #  cdf(value = 0.08) <= u < cdf(value = 0.5)
        elif u < self.kroupa_cdf(self._min_value, 0.5, self._norm):
            if self._min_value < 0.08:
                value = (((u - (self._norm / 0.7 * 0.08**0.3 *
                         (0.08**0.7 - self._min_value**0.7))) *
                        (-0.3) / self._norm * 0.08**(-1.3) +
                    0.08**(-0.3))**(1 / -0.3))
            else:  # 0.08 <= minvalue < 0.5
                value = (u * (-0.3 / self._norm) * 0.08**(-1.3) +
                         self._min_value**(-0.3))**(1/-0.3)
        #  cdf(value = 0.5) <= u
        else:
            if self._min_value < 0.08:
                value = ((u - (self._norm / 0.7 * 0.08**0.3 *
                         (0.08**0.7 - self._min_value**0.7)) -  # - CDF 0
                        ((self._norm / -0.3) * 0.08**1.3 *
                         (0.5**(-0.3) - 0.08**(-0.3)))) *  # - CDF 1
                         -1.3/self._norm * 0.08**(-1.3) * 2 +
                         0.5**(-1.3))**(1/-1.3)  # /relevant part of CDF 2

            elif self._min_value < 0.5:  # 0.08 <= self._min_value < 0.5
                value = ((u - ((self._norm / -0.3) * 0.08**1.3 *
                         (0.5**(-0.3) - self._min_value**(-0.3)))
                    ) * (-1.3/self._norm) * 0.08**(-1.3) * 2 +
                    0.5**(-1.3))**(1/-1.3)
            else:  # 0.5 <= self._min_value
                value = (u * (-1.3/self._norm) * 0.08**(-1.3) *
                         2 + self._min_value**(-1.3))**(1/-1.3)

        value = (value - self._min_value) / (self._max_value - self._min_value)
        # np.clip in case of python errors in line above
        return np.clip(value, 0.0, 1.0)



"""Definitions for the `AsphericalKilonova` class."""
import numpy as np
from astrocats.catalog.source import SOURCE

from mosfit.modules.energetics.energetic import Energetic


# Important: Only define one ``Module`` class per file.


class AsphericalKilonova(Energetic):
    """
        Uses viewing angle and lanthanide-poor opening angle from Darbha and
        Kasen 2020.

        NOTE only valid for theta_open <= pi/2
    """

    _REFERENCES = [
        {SOURCE.BIBCODE: '2020ApJ...897..150D'}
    ]

    def process(self, **kwargs):
        """Process module."""
        # Viewing angle
        self._cos_theta = kwargs[self.key('cos_theta')]
        # Opening angle
        self._cos_theta_open = kwargs[self.key('cos_theta_open')]

        # Dharba and Kasen 2020 scalings for conical opening in opaque sphere

        ct = (1-self._cos_theta_open**2)**0.5

        if self._cos_theta > ct:
            Aproj_top = np.pi * ct * self._cos_theta
        else:
            theta_p = np.arccos(self._cos_theta_open /
                                (1 - self._cos_theta**2)**0.5)
            theta_d = np.arctan(np.sin(theta_p) / self._cos_theta_open *
                        (1 - self._cos_theta**2)**0.5 / np.abs(self._cos_theta))
            Aproj_top = (theta_p - np.sin(theta_p)*np.cos(theta_p)) - (ct *
                            self._cos_theta*(theta_d -
                            np.sin(theta_d)*np.cos(theta_d) - np.pi))

        minus_cos_theta = -1 * self._cos_theta

        if minus_cos_theta < -1 * ct:
            Aproj_bot = 0
        else:
            theta_p2 = np.arccos(self._cos_theta_open /
                                (1 - minus_cos_theta**2)**0.5)
            theta_d2 = np.arctan(np.sin(theta_p2) / self._cos_theta_open *
                        (1 - minus_cos_theta**2)**0.5 / np.abs(minus_cos_theta))

            Aproj_bot1 = (theta_p2 - np.sin(theta_p2)*np.cos(theta_p2)) + (ct *
            minus_cos_theta*(theta_d2 - np.sin(theta_d2)*np.cos(theta_d2)))
            Aproj_bot = np.max([Aproj_bot1, 0])

        Aproj = Aproj_top + Aproj_bot


        # Compute reference areas for this opening angle to scale luminosity

        cos_theta_ref = 0.5

        if cos_theta_ref > ct:
            Aref_top = np.pi * ct * cos_theta_ref
        else:
            theta_p_ref = np.arccos(self._cos_theta_open /
                                (1 - cos_theta_ref**2)**0.5)
            theta_d_ref = np.arctan(np.sin(theta_p_ref) / self._cos_theta_open *
                        (1 - cos_theta_ref**2)**0.5 / np.abs(cos_theta_ref))
            Aref_top = (theta_p_ref - np.sin(theta_p_ref) *
                        np.cos(theta_p_ref)) - (ct * cos_theta_ref *
                        (theta_d_ref - np.sin(theta_d_ref) *
                        np.cos(theta_d_ref) - np.pi))

        minus_cos_theta_ref = -1 * cos_theta_ref

        if minus_cos_theta_ref < -1 * ct:
            Aref_bot = 0
        else:
            theta_p2_ref = np.arccos(self._cos_theta_open /
                                (1 - minus_cos_theta_ref**2)**0.5)
            theta_d2_ref = np.arctan(np.sin(theta_p2_ref) /
                    self._cos_theta_open * (1 - minus_cos_theta_ref**2)**0.5 /
                        np.abs(minus_cos_theta_ref))

            Aref_bot = (theta_p2_ref - np.sin(theta_p2_ref) *
                        np.cos(theta_p2_ref)) + (ct * minus_cos_theta_ref *
                        (theta_d2_ref - np.sin(theta_d2_ref) *
                        np.cos(theta_d2_ref)))

        Aref = Aref_top + Aref_bot

        Ablue = Aproj
        Ablue_ref = Aref

        Ared = np.pi - Ablue
        Ared_ref = np.pi - Ablue_ref


        return {self.key('area_blue'): Ablue,
                self.key('area_blue_ref'): Ablue_ref,
                self.key('area_red'): Ared,
                self.key('area_red_ref'): Ared_ref
                }



"""Definitions for the `BNSConstraints` class."""
import astropy.constants as c
import numpy as np

from mosfit.constants import C_CGS, M_SUN_CGS, KM_CGS, G_CGS
from mosfit.modules.constraints.constraint import Constraint

# G_CGS = c.G.cgs.value


class BNSConstraints(Constraint):
    """BNS constraints.

    1. M1 < Mtov
    2. M2 > 0.8 Msun
    3. 9 < R_ns < 16
    4. Causality relation relating Mtov and radius

    Free parameters are Mchirp and q. Thus for large q, M1 can be larger or M2
    smaller than allowed by NS EoS. Constraint penalises masses outside range

    Realistic EoS prevents extreme values of radius
    """

    def __init__(self, **kwargs):
        """Initialize module."""
        super(BNSConstraints, self).__init__(**kwargs)

    def process(self, **kwargs):
        """Process module. Add constraints below."""
        self._score_modifier = 0.0
        # Mass of heavier NS
        self._m1 = kwargs[self.key('M1')]
        # Mass of lighter NS
        self._m2 = kwargs[self.key('M2')]
        self._m_tov = kwargs[self.key('Mtov')]
        self._r1 = kwargs[self.key('R1')]
        self._r2 = kwargs[self.key('R2')]

        # Soft max/min, proportional to diff^2 and scaled to -100 for 0.1 Msun
        # 1
        if self._m1 > self._m_tov:
            self._score_modifier -= (100. * (self._m1-self._m_tov))**2

        # 2
        if self._m2 < 0.8:
            self._score_modifier -= (100. * (0.8-self._m2))**2

        # 3
        if self._r1 > 16:
            self._score_modifier -= (20. * (self._r1-16))**2

        if self._r1 < 9:
            self._score_modifier -= (20. * (9-self._r1))**2

        if self._r2 > 16:
            self._score_modifier -= (20. * (self._r2-16))**2

        if self._r2 < 9:
            self._score_modifier -= (20. * (9-self._r2))**2


        # 4
        Mcaus = 1/2.82 * C_CGS**2 * self._r1 * KM_CGS / G_CGS / M_SUN_CGS

        if self._m_tov > Mcaus:
            self._score_modifier -= (100. * (self._m_tov-Mcaus))**2

        return {self.key('score_modifier'): self._score_modifier}



"""List of numerical constants, faster than using astropy with every call."""

from decimal import Decimal

import astropy.constants as c
import astropy.units as u
import numpy as np

LIKELIHOOD_FLOOR = -np.inf
LOCAL_LIKELIHOOD_FLOOR = -1.0e8

ANG_CGS = u.Angstrom.cgs.scale
AU_CGS = u.au.cgs.scale
C_CGS = c.c.cgs.value
H_CGS = c.h.cgs.value
DAY_CGS = u.day.cgs.scale
FOE = 1.0e51
FOUR_PI = 4.0 * np.pi
SQRT_2_PI = np.sqrt(2.0 * np.pi)
IPI = 1.0 / np.pi
KM_CGS = u.km.cgs.scale
M_SUN_CGS = c.M_sun.cgs.value
M_P_CGS = c.m_p.cgs.value
MEV_CGS = u.MeV.cgs.scale
MAG_FAC = 2.5
MPC_CGS = u.Mpc.cgs.scale

KS_DAYS = float(Decimal('1000') / Decimal(DAY_CGS))
H_C_CGS = H_CGS * C_CGS
H_C_ANG_CGS = H_C_CGS / ANG_CGS

G_CGS = c.G.cgs.value
